SuperStrict
'Import "perfect_snake_api.c"		'easier version, bmk starts gcc
Import "perfect_snake_api.o"		'better choice if you started gcc manually (to pass option -O3)

Extern

	Function getPathLength:Int()
	Function getPath(p:Byte Ptr)
	Function initSearch(fieldSize:Int, initial:Byte Ptr, initialSize:Int, finalPath:Byte Ptr, finalSize:Int)
	Function continueSearch:Int(iterations:Int)
	Function hasFinished:Int()
	Function getNoOfPerfects:Int()
	Function resetPerfects()
	Function resetSearch()
	
End Extern

Const N:Int = 0, E:Int = 1, S:Int = 2, W:Int = 3
Global vx:Int[] = [0, 1, 0, -1], vy:Int[] = [-1, 0, 1, 0]
Global fieldSize:Int

Function CreateStep:TStep(x:Int, y:Int, dir:Int)

	Local s:TStep = New TStep
	s.x = x
	s.y = y
	s.dir = dir
	Return s
	
End Function

Type TStep

	Field x:Int, y:Int, dir:Int

End Type

Function PathToArray:Byte[](path:TStep[])

	Local a:Byte[path.length]
	For Local i:Int = 0 Until path.length
	
		a[i] = path[i].dir
		
	Next
	Return a
	
End Function

Function PathFromArray:TStep[](a:Byte[])

	Local path:TStep[a.length]
	Local x:Int = 1, y:Int = 0
	For Local i:Int = 0 Until a.length
	
		Local d:Int = a[i]
		x :+ vx[d]
		y :+ vy[d]
		path[i] = CreateStep(x, y, d)
		
	Next
	
	Return path
	
End Function

Function ArrayFromString:Byte[](s$)

	Local a:Int[255]
	a[Asc("N")] = N
	a[Asc("E")] = E
	a[Asc("S")] = .S
	a[Asc("W")] = W
	Local ar:Byte[s.length]
	For Local i:Int = 0 Until s.length
	
		ar[i] = a[s[i]]
		
	Next
	Return ar
	
End Function

Function StringFromArray$(a:Byte[])

	Local ar$[] = ["N", "E", "S", "W"]
	Local s$
	For Local i:Int = 0 Until a.length
	
		s :+ ar[a[i]]
		
	Next
	Return s
	
End Function

Function PathIsValid:Int(p:TStep[])

	Local prev:TStep = CreateStep(1, 0, N)
	
	For Local s:TStep = EachIn p
	
		If s.x<0 Or s.y<0 Or s.x>=fieldSize Or s.y>=fieldSize Then Return 0
		If s.x <> prev.x+vx[s.dir] Or s.y <> prev.y+vy[s.dir] Then Return 0
		prev = s
		
	Next
	Return 1
	
End Function

Function GetNextPath:TStep[](path:TStep[])	'assuming path is invalid

	If path.length<2 Then Return Null
	Local dir:Int = path[path.length-1].dir
	path = path[..path.length-1]
	dir :+ 1
	If dir >= 4 Then Return GetNextPath(path)
	Local last:TStep = path[path.length-1]
	Local x:Int = last.x+vx[dir], y:Int = last.y+vy[dir]
	path :+ [CreateStep(x,y,dir)]
	If (x=0 Or x=1) And y=0 Then Return GetNextPath(path) 
	For Local s:TStep = EachIn path
	
		If s.x = x And s.y = y Then Return GetNextPath(path)
		
	Next
	If Not PathIsValid(path) Then Return GetNextPath(path)
	Return path
	
End Function

Function GetPathPosition!(a:Byte[])

	If a.length=0 Then Return 0
	If a[0] <> E And a[0] <> S Then Return (a[0] = W)
	Local add!
	If a[0] = S Then add = 0.5
	Local value!, factor! = 1.0/6.0
	Const EPSILON! = 0.0001
	Local from:Int = (a[0]+2) Mod 4
	For Local i:Int = 1 Until a.length

		value :+ (a[i]-(a[i]>from)) * factor
		factor :* Double(1.0)/Double(3.0)
		
		If factor<EPSILON Then Exit
		
	Next
	Return value+add
	
End Function

Function CleanupPath$(path$)

	Local clean$
	For Local i:Int = 0 Until path.length
	
		Select Chr(path[i])
		
			Case "N", "E", "S", "W"
			
				clean :+ Chr(path[i])
				
		End Select
		
	Next
	Return clean
	
End Function

If AppArgs.length=0 Then End

If AppArgs.length<4 Then

	Print "Usage: "+AppArgs[0]+" [field size] [start path] [end path]"
	Print "ATTENTION: No error checking is done!"
	End
	
EndIf

Global iterationsPerUnit! = 2000000
Const TIME_UNIT! = 900.0	'Change this value to adjust CPU usage
													'It determines, how many milliseconds of a second should be spent on computation
													'The default value of 900.0 means a CPU usage of 90%.
Global initialPath:TStep[], finalPath:TStep[]

fieldSize = Int(AppArgs[1])

initialPath = PathFromArray(ArrayFromString(AppArgs[2]))
finalPath   = PathFromArray(ArrayFromString(AppArgs[3]))
If InitializeSearch() Then End

Local timer:TTimer = CreateTimer(1)

Local cnt:Int = 0
Local totalTime:Int
Local startPath$ = AppArgs[2]

Repeat

	WaitTimer timer

	Local iterations:Int = Int(iterationsPerUnit)
	Local start:Int = MilliSecs()
	Local result:Int = continueSearch(iterations)
	Local time:Int = MilliSecs()-start
	totalTime :+ time
	cnt :+ 1
	If cnt>=10 Or result Then
	
		Local perfects:Int = getNoOfPerfects()
		resetPerfects
		Local endPath$
		If result Then
		
			endPath = AppArgs[3]
			
		Else
		
			Local currentPath:Byte[] = GetCurrentPath()
			endPath = StringFromArray(currentPath)
			
		EndIf
		Print startPath+";"+endPath+";"+perfects+";"+totalTime
		totalTime=0
		
		cnt=0
		startPath = endPath
		
	EndIf
	
	If result Then End
	iterationsPerUnit = 0.9*iterationsPerUnit + 0.1*Double(iterations)*TIME_UNIT/Double(time)

Forever

Global ipa:Byte[], fpa:Byte[]

Function InitializeSearch:Int()

	Repeat
	
		ipa = PathToArray(initialPath)
		fpa = PathToArray(finalPath)
		
		Local res:Int = initSearch(fieldSize, ipa, ipa.length, fpa, fpa.length)
		If res = 1 Then
		
			Return 1
			
		ElseIf res=-1 Then
		
			initialPath = GetNextPath(initialPath)
			If Not initialPath Then
			
				Return 1
				
			EndIf
			
		Else
		
			Exit
			
		EndIf
		
	Forever
	
End Function

Function GetCurrentPath:Byte[]()

	Local a:Byte[getPathLength()]
	getPath a
	Return a
	
End Function
