SuperStrict
Import MaxGUI.Drivers
Import pub.freeprocess

Const N:Int = 0, E:Int = 1, S:Int = 2, W:Int = 3
Global vx:Int[] = [0, 1, 0, -1], vy:Int[] = [-1, 0, 1, 0]
Global fieldSize:Int

Function CreateStep:TStep(x:Int, y:Int, dir:Int)

	Local s:TStep = New TStep
	s.x = x
	s.y = y
	s.dir = dir
	Return s
	
End Function

Type TStep

	Field x:Int, y:Int, dir:Int

End Type

Function PathToArray:Byte[](path:TStep[])

	Local a:Byte[path.length]
	For Local i:Int = 0 Until path.length
	
		a[i] = path[i].dir
		
	Next
	Return a
	
End Function

Function PathFromArray:TStep[](a:Byte[])

	Local path:TStep[a.length]
	Local x:Int = 1, y:Int = 0
	For Local i:Int = 0 Until a.length
	
		Local d:Int = a[i]
		x :+ vx[d]
		y :+ vy[d]
		path[i] = CreateStep(x, y, d)
		
	Next
	
	Return path
	
End Function

Function ArrayFromString:Byte[](s$)

	Local a:Int[255]
	a[Asc("N")] = N
	a[Asc("E")] = E
	a[Asc("S")] = .S
	a[Asc("W")] = W
	Local ar:Byte[s.length]
	For Local i:Int = 0 Until s.length
	
		ar[i] = a[s[i]]
		
	Next
	Return ar
	
End Function

Function StringFromArray$(a:Byte[])

	Local ar$[] = ["N", "E", "S", "W"]
	Local s$
	For Local i:Int = 0 Until a.length
	
		s :+ ar[a[i]]
		
	Next
	Return s
	
End Function

Function PathIsValid:Int(p:TStep[])

	Local prev:TStep = CreateStep(1, 0, N)
	
	For Local s:TStep = EachIn p
	
		If s.x<0 Or s.y<0 Or s.x>=fieldSize Or s.y>=fieldSize Then Return 0
		If s.x <> prev.x+vx[s.dir] Or s.y <> prev.y+vy[s.dir] Then Return 0
		prev = s
		
	Next
	Return 1
	
End Function

Function GetPathPosition!(a:Byte[])

	If a.length=0 Then Return 0
	If a[0] <> E And a[0] <> S Then Return (a[0] = W)
	Local add!
	If a[0] = S Then add = 0.5
	Local value!, factor! = 1.0/6.0
	Const EPSILON! = 0.0001
	Local from:Int = (a[0]+2) Mod 4
	For Local i:Int = 1 Until a.length

		value :+ (a[i]-(a[i]>from)) * factor
		factor :* Double(1.0)/Double(3.0)
		
		If factor<EPSILON Then Exit
		
	Next
	Return value+add
	
End Function

Function CleanupPath$(path$)

	Local clean$
	For Local i:Int = 0 Until path.length
	
		Select Chr(path[i])
		
			Case "N", "E", "S", "W"
			
				clean :+ Chr(path[i])
				
		End Select
		
	Next
	Return clean
	
End Function

Function ComparePaths:Int(a:Byte[], b:Byte[])

	For Local i:Int = 0 Until Min(a.length,b.length)
	
		If a[i]<b[i] Then Return -1
		If a[i]>b[i] Then Return 1
		
	Next
	If a.length<b.length Then Return -1
	If a.length>b.length Then Return 1
	Return 0
	
End Function

Function UpdatePositions(gad:TGadget)

	Local txt$ = CleanupPath(TextFieldText(gad).ToUpper())
	Local val!
	If txt Then
	
		val = GetPathPosition(ArrayFromString(txt))
		
	Else
	
		val = -1
		
	EndIf
	Local ip:Byte[], fp:Byte[]
	Select gad
	
		Case inputInitial
		
			currentWorker.startValue = val
			ip = ArrayFromString(txt)
			fp = ArrayFromString(CleanupPath(TextFieldText(inputFinal).ToUpper()))
			
		Case inputFinal
		
			currentWorker.endValue = val
			fp = ArrayFromString(txt)
			ip = ArrayFromString(CleanupPath(TextFieldText(inputInitial).ToUpper()))
			
	End Select
	

	If currentWorker.startValue <> -1 And currentWorker.endValue <> -1 And ComparePaths(ip,fp) <= 0 Then
	
		Local x# = currentWorker.startValue, w# = (currentWorker.endValue-currentWorker.startValue)
		x :* GadgetWidth(panelPosition)
		w :* GadgetWidth(panelPosition)
	
		If innerPanel Then
		
			SetGadgetShape innerPanel, x, 0, w, 20
			
		Else
		
			innerPanel = CreatePanel(x, 0, w, 20, panelPosition, PANEL_RAISED)
			SetGadgetColor innerPanel, 0, 255, 0
			SetGadgetLayout innerPanel, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE
			EnableGadget buttonStart
			
		EndIf
		
	ElseIf innerPanel Then
	
		FreeGadget innerPanel
		innerPanel = Null
		DisableGadget buttonStart
		
	EndIf
	
End Function

Const CONFIG_FILE$ = "saw.ini"


Global logDir$ = "log"
Global parallelProcesses:Int = 1
Global maxLines:Int = 10000
Global crashFile$ = "~~"+StripAll(AppFile)+".temp"
Local stream:TStream = ReadFile(CONFIG_FILE)
While stream And Not stream.Eof()
	
	Local line$ = stream.ReadLine().Trim()
	If line.StartsWith("#") Then Continue
	Local pos:Int = line.Find("=")
	If pos=-1 Then Continue
	Local val$ = line[pos+1..].Trim()
	Select line[..pos].Trim().ToLower()
	
		Case "logdir"
		
			logDir = val
						
		Case "parallelprocesses"
		
			parallelProcesses = Max(1, Int(val))
			
		Case "maxlines"
		
			maxLines = Max(1, Int(val))
			
		Case "crashfile"
		
			crashFile = val
			
	End Select
	
Wend

If Not stream Then

	stream = WriteFile(CONFIG_FILE)
	stream.WriteLine("# config file for "+AppFile)
	stream.WriteLine("# delete this file to restore default settings")
	stream.WriteLine("# directory to save log files (see 'logDir'/toc.txt for an overview)")
	stream.WriteLine("logDir = "+logDir)
	stream.WriteLine("# number of parallel processes (best to be the number of CPU kernels you want to use")
	stream.WriteLine("parallelProcesses = "+parallelProcesses)
	stream.WriteLine("# the maximum number of lines per log file")
	stream.WriteLine("maxLines = "+maxLines)
	stream.WriteLine("# a file, which will be created on application start and deleted on proper termination")
	stream.WriteLine("# enabling the application to discover its crashs and report them")
	stream.WriteLine("crashFile = "+crashFile)
	
EndIf

logDir = StripSlash(logDir)+"/"

Select FileType(logDir)

	Case 0
	
		If Not CreateDir(logDir) Then RuntimeError "Failed to create '"+logDir+"'"
		
	Case 2
		
	Default
	
		RuntimeError "'"+logDir+"' is not a directory!"
		
End Select
	
	

stream.Close
Global tocFile$ = logDir+"toc.txt"
Global tocStream:TStream = OpenStream(tocFile)
If Not tocStream Then tocStream = WriteFile(tocFile)
SeekStream tocStream, StreamSize(tocStream)

Select FileType(crashFile)

	Case 1
	
		Local str$ = "A crash of the program started "+LoadText(crashFile).Trim()+" has been discovered~r~n"+..
									"This error also might be caused by parallel executition of two instances of this application"
		tocStream.WriteLine CurrentDate()+" "+CurrentTime()+": "+str.Replace("~n", "~n   ")
		Notify str+"~r~nThe error has been logged in "+tocFile
		
	Case 2
	
		RuntimeError "crashfile '"+crashFile+"' is a directory!"
		
End Select

stream = WriteFile(crashFile)
stream.WriteLine CurrentDate()+" "+CurrentTime()
stream.Close
stream=Null


Function Cleanup()

	tocStream.Close
	tocStream = Null
	DeleteFile crashFile
	
End Function
OnEnd Cleanup

Function GetFreeLogFile$()

	Local i:Int = 1
	Repeat
	
		Local name$ = logDir+"log"+i+".txt"
		If FileType(name) = 0 Then Return name
		i :+ 1
		
	Forever
	
End Function


Function CreateWorker:TWorker()

	Local w:TWorker = New TWorker
	w.index = w.array.length
	w.name = w.cnt
	w.cnt :+ 1
	w.array :+ [w]
	AddGadgetItem workerTabs, w.name
	
	EnableGadget buttonRemoveWorker
	
	Return w
	
End Function
	


Type TWorker

	Global array:TWorker[]
	Global cnt:Int=1
	
	Field index:Int, name:Int
	Field initialPath$, finalPath$
	Field running:Int = False, started$
	Field logFiles$[]
	Field currentLog:TStream, lines:Int = 0
	Field startValue!, endValue!
	Field perfectsTotal:Long = 0
	Field discardSteps:Int, initialValue!, finalValue!
	Field currentPath$
	Field process:TProcess
	
	Method Remove()
	
		If running Then Stop
		RemoveGadgetItem workerTabs, index
		For Local i:Int = index+1 Until array.length
		
			array[i].index :- 1
			
		Next
		array = array[..index]+array[index+1..]
		If array.length=1 Then DisableGadget buttonRemoveWorker
		
	End Method
	
	Method Stop()
	
		StopRun "stopped"
		
	End Method
	
	Method StopRun(comment$)
	
		DebugLog "StopRun("+comment+")"
	
		If process.Status() Then
		
			Update
			
		Else
		
			RestoreWindow win
			Focus
			
		EndIf
		OutputLine "# "+comment
		currentLog.Close
		currentLog = Null
		Local tocEntry$ = started+" - "+(CurrentDate()+" "+CurrentTime())+" "+..
											initialPath+"-"+currentPath+" ("+(", ".Join(logFiles))+") ["+comment+"]: "+perfectsTotal
		tocStream.WriteLine tocEntry
		running=False
		process.Terminate
		process=Null
		
		If currentWorker=Self Then
		
			FillGUI
			
		EndIf
		
	End Method
	
	Method Update()
	
		If running Then
		
			If Not process.Status() Then
			
				Repeat
				
					Local line$ = process.pipe.ReadLine()
					If Not line Then Exit
					OutputLine line
					Local parts$[] = line.Split(";")
					If parts.length<>4 Then
					
						OutputLine("# ???")
						
					Else
					
						perfectsTotal :+ Long(parts[2])
						
					EndIf
					
				Forever
				
				currentPath=finalPath
	
				StopRun "terminated"
				Focus
				RestoreWindow win
				Return
				
			EndIf
			
			Local lastLine$
			Local parts$[]
			Repeat
			
				Local line$ = process.pipe.ReadLine()
				If Not line Then Exit
				OutputLine line
				lastLine = line
				parts = line.Split(";")
				If parts.length<>4 Then
				
					OutputLine("# ???")
					
				Else
				
					perfectsTotal :+ Long(parts[2])
					
				EndIf
				
			Forever

			If parts.length<>4 Then Return
			
			currentPath = parts[1]
			If currentWorker=Self Then UpdateUI
	
			
		EndIf
		
	End Method
		
	Method UpdateUI()
	
		Local val! = GetPathPosition(ArrayFromString(currentPath[discardSteps..]))
		If val < initialValue Or val > finalValue Then
		
			UpdateProgBar progbar, 0
			
		Else
		
			Local v! = (val-initialValue)/(finalValue-initialValue)
			UpdateProgBar progbar, v
			
		EndIf
		
		SetGadgetText inputPerfects, perfectsTotal
		SetGadgetText inputCurrentPath, currentPath
		
	End Method
	
	Method FillGUI()
	
		SetGadgetText inputInitial, initialPath
		SetGadgetText inputFinal, finalPath
		SetGadgetText inputPerfects, perfectsTotal
		SetGadgetText inputCurrentPath, currentPath
		EnableGadget buttonStart
		UpdatePositions inputInitial
		If running Then
		
			UpdateUI
			DisableGadget inputInitial
			DisableGadget inputFinal
			DisableGadget buttonStart
			EnableGadget progbar
			EnableGadget buttonStop
			
		Else
		
			DisableGadget buttonStop
			UpdateProgBar progbar, 0
			DisableGadget progbar
			EnableGadget inputInitial
			EnableGadget inputFinal
			
		EndIf
		
	End Method
	
	Method OutputLine(line$)
	
		If lines>=maxLines Then
		
			lines=0
			currentLog.Close
			Local file$ = GetFreeLogFile()
			logFiles :+ [file]
			currentLog = WriteFile(file)
			
		EndIf
		lines :+ 1
		currentLog.WriteLine line
		
	End Method
	
	Method Focus()
	
		SelectGadgetItem workerTabs, index
		FillGUI
		
	End Method
	
	Method START()

		initialPath = CleanupPath(TextFieldText(inputInitial))
		finalPath   = CleanupPath(TextFieldText(inputFinal))
		process = CreateProcess(".saw_worker.exe "+fieldSize+" "+initialPath+" "+finalPath)
		For discardSteps = 0 Until Min(initialPath.length, finalPath.length)
		
			If initialPath[discardSteps] <> finalPath[discardSteps] Then Exit
			
		Next
		initialValue = GetPathPosition(ArrayFromString(initialPath[discardSteps..]))
		finalValue   = GetPathPosition(ArrayFromString(  finalPath[discardSteps..]))
		perfectsTotal = 0
		
		Local file$ = GetFreeLogFile()
		logFiles = [file]
		currentLog = WriteFile(file)
		lines=0
		currentPath = initialPath
		running=True
		If currentWorker=Self Then
		
			FillGUI
			
		EndIf
		started = CurrentDate()+" "+CurrentTime()
		
	End Method
	
End Type
	



Global win:TGadget = CreateWindow("self avoiding walk", 0, 0, 300, 260, , ..
												WINDOW_CENTER|WINDOW_CLIENTCOORDS|WINDOW_TITLEBAR|WINDOW_RESIZABLE)
Global workerTabs:TGadget = CreateTabber(10, 10, 280, 25, win)
Local lab1:TGadget = CreateLabel("initial:", 10, 40, 70, 20, win)
Global inputInitial:TGadget = CreateTextField(90, 40, 200, 20, win)
Local lab2:TGadget = CreateLabel("final:", 10, 70, 70, 20, win)
Global inputFinal:TGadget = CreateTextField(90, 70, 200, 20, win)
Global panelPosition:TGadget = CreatePanel(10, 100, 150, 20, win, PANEL_SUNKEN)
Global buttonStart:TGadget = CreateButton("START!", 170, 100, 50, 20, win)
Global buttonStop :TGadget = CreateButton("STOP!", 240, 100, 50, 20, win)
Global progbar:TGadget = CreateProgBar(10, 130, 280, 20, win)
Local lab3:TGadget = CreateLabel("current path: ", 10, 160, 70, 20, win)
Global inputCurrentPath:TGadget = CreateTextField(90, 160, 200, 20, win)
Local lab4:TGadget = CreateLabel("# perfects: ", 10, 190, 70, 20, win)
Global inputPerfects:TGadget = CreateTextField(90, 190, 200, 20, win)

Global buttonAddWorker:TGadget = CreateButton("add worker", 10, 230, 135, 20, win)
Global buttonRemoveWorker:TGadget = CreateButton("remove worker", 155, 230, 135, 20, win)

Global innerPanel:TGadget


DisableGadget buttonStop
UpdateProgBar progbar, 0
DisableGadget progbar
DisableGadget buttonStart

For Local gad:TGadget = EachIn [workerTabs, lab1, inputInitial, lab2, inputFinal, panelPosition, buttonStart,..
			buttonStop, progbar, lab3, inputCurrentPath, lab4, inputPerfects, buttonAddWorker, buttonRemoveWorker]

	SetGadgetLayout gad, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE, EDGE_RELATIVE
	
Next


Function MyFilter:Int(event:TEvent, context:Object)

	If event.id <> EVENT_KEYCHAR Then Return 1
	
	Select Chr(event.data)
	
		Case "N", "E", "S", "W", "n", "e", "s", "w"
		'Case KEY_N, KEY_E, KEY_S, KEY_W, KEY_BACKSPACE, KEY_DELETE, KEY_LEFT, KEY_RIGHT, KEY_RETURN, KEY_ESCAPE
		
			Return True
			
	End Select
	If event.data < 32 Then Return True
	Return False
	
End Function

SetGadgetFilter inputInitial, MyFilter
SetGadgetFilter inputFinal  , MyFilter


Global currentWorker:TWorker = CreateWorker()
DisableGadget buttonRemoveWorker


Global startValue!=-1, endValue!=-1

fieldSize = 9

Global ipa:Byte[], fpa:Byte[]
Global initialPath:TStep[], finalPath:TStep[]

Global process:TProcess

Global discardSteps:Int = 0, initialValue!, finalValue!
Global perfectsTotal:Long

Global timer:TTimer = CreateTimer(1)

Repeat

	Select WaitEvent()
	
		Case EVENT_WINDOWCLOSE, EVENT_APPTERMINATE
		
			For Local w:TWorker = EachIn TWorker.array
			
				If w.running Then w.StopRun "closed"
				
			Next
			End
			
		Case EVENT_GADGETACTION
		
			Select CurrentEvent.source
			
				Case inputInitial
				
					currentWorker.initialPath = TextFieldText(inputInitial).ToUpper()
					SetGadgetText inputInitial, currentWorker.initialPath
					UpdatePositions inputInitial
					
				Case inputFinal
				
					currentWorker.finalPath = TextFieldText(inputFinal).ToUpper()
					SetGadgetText inputFinal, currentWorker.finalPath
					SetGadgetText inputFinal, TextFieldText(inputFinal).ToUpper()
					UpdatePositions inputFinal
					
				Case buttonStop
				
					currentWorker.Stop
					
				Case buttonStart
				
					currentWorker.Start
					
				Case workerTabs
				
					currentWorker = TWorker.array[CurrentEvent.data]
					currentWorker.FillGUI
					
				Case buttonAddWorker
				
					CreateWorker().Focus
					
				Case buttonRemoveWorker
				
					If currentWorker.running Then currentWorker.StopRun "closed"
					Local index:Int = currentWorker.index
					currentWorker.Remove
					If index>=TWorker.array.length Then index = TWorker.array.length-1
					currentWorker = TWorker.array[index]
					currentWorker.Focus
					
			End Select
			
		Case EVENT_TIMERTICK
		
			For Local w:TWorker = EachIn TWorker.array
			
				w.Update
				
			Next
		
	End Select
Forever
End
